"""
Helpers for bboxes preprocessing

"""
import pandas as pd
import re
import numpy as np
from typing import List, Tuple, Optional, Any


def expand_bbox(x: np.array) -> np.array:
    r = np.array(re.findall("([0-9]+[.]?[0-9]*)", x))
    if len(r) == 0:
        r = [-1, -1, -1, -1]

    return r


def preprocess_boxes(train_boxes_df: pd.DataFrame) -> pd.DataFrame:

    (
        train_boxes_df["x"],
        train_boxes_df["y"],
        train_boxes_df["w"],
        train_boxes_df["h"],
    ) = (-1, -1, -1, -1)
    train_boxes_df[["x", "y", "w", "h"]] = np.stack(
        train_boxes_df["bbox"].apply(lambda x: expand_bbox(x))
    )
    train_boxes_df.drop(columns=["bbox"], inplace=True)
    train_boxes_df["x"] = train_boxes_df["x"].astype(np.float)
    train_boxes_df["y"] = train_boxes_df["y"].astype(np.float)
    train_boxes_df["w"] = train_boxes_df["w"].astype(np.float)
    train_boxes_df["h"] = train_boxes_df["h"].astype(np.float)

    return train_boxes_df


def filter_box_size(
    train_boxes_df: pd.DataFrame,
    min_size: Optional[int] = None,
    max_size: Optional[int] = 700,
) -> pd.DataFrame:
    """
    Apply filtering for boxes by size
        Args:
            boxes_df: pd.DataFrame with train boxes coordinates
            min_size: boxes with the h, w below minimum are removed
            max_size: boxes with the h, w above maximum are removed

        Output:
            pd.DataFrame with filtered boxes
    """
    train_boxes_df["area"] = train_boxes_df["w"] * train_boxes_df["h"]
    if min_size:
        size_filter = train_boxes_df["w"] > min_size & train_boxes_df["h"] > min_size
        train_boxes_df = train_boxes_df[size_filter]
    if max_size:
        size_filter = train_boxes_df["w"] < max_size & train_boxes_df["h"] < max_size
        train_boxes_df = train_boxes_df[size_filter]

    return train_boxes_df


def filter_box_area(
    train_boxes_df: pd.DataFrame,
    min_area: Optional[int] = 10,
    max_area: Optional[int] = 200000,
) -> pd.DataFrame:
    """
    Apply filtering for boxes by area
        Args:
            boxes_df: pd.DataFrame with train boxes coordinates
            min_area: boxes with the area below minimum are removed
            max_area: boxes with the area above maximum are removed

        Output:
            pd.DataFrame with filtered boxes
    """
    train_boxes_df["area"] = train_boxes_df["w"] * train_boxes_df["h"]
    if min_area:
        area_filter = train_boxes_df["area"] > min_area
        train_boxes_df = train_boxes_df[area_filter]
    if max_area:
        area_filter = train_boxes_df["area"] < max_area
        train_boxes_df = train_boxes_df[area_filter]

    return train_boxes_df


def get_boxes(boxes_df: pd.DataFrame, image_id: str) -> np.array:
    records = boxes_df[boxes_df["image_id"] == image_id]
    return records[["x", "y", "w", "h"]].values


def split_prediction_string(preds: str) -> Tuple[np.array, np.array]:
    """
    Split string predictions into bboxes coordinates (locations) and scores
    for bboxes predictions
    """
    parts = preds.split()
    if len(parts) % 5 != 0:
        raise ValueError("Predictions should have lens of 5: 4 coords and a label")

    locations, scores = [], []
    for ind in range(len(parts) // 5):
        scores.append(float(parts[ind * 5]))
        location = (
            int(float(parts[ind * 5 + 1])),
            int(float(parts[ind * 5 + 2])),
            int(float(parts[ind * 5 + 3])),
            int(float(parts[ind * 5 + 4])),
        )
        locations.append(np.array(location))
    if len(locations) <= 0:
        raise ValueError("No locations")

    return np.array(locations), np.array(scores)


def format_prediction_string(boxes: List[Any], scores: List[Any]) -> str:
    """
    Creates a string of bboxes preditions: confidence scores and coordinates
    """
    pred_strings = []
    for j in zip(scores, boxes):
        pred_strings.append(
            "{0:.4f} {1} {2} {3} {4}".format(j[0], j[1][0], j[1][1], j[1][2], j[1][3])
        )

    return " ".join(pred_strings)



"""
BCE - Jaccard Loss

Adapted from:
    https://github.com/BloodAxe/pytorch-toolbelt/blob/develop/pytorch_toolbelt/losses/
    Author: E. Khvedchenya (BloodAxe)
    and https://github.com/ternaus/robot-surgery-segmentation/blob/master/loss.py
    Author: V.Iglovikov (ternaus)
"""
import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import BCEWithLogitsLoss
from torch.nn.modules.loss import _Loss

__all__ = ["BCEJaccardLoss"]


class BCEJaccardLoss(_Loss):
    """
    Implementation of combination of BCE and soft Jaccard loss
    for binary image segmentation task, calculated from logits

    """

    def __init__(
        self,
        bce_weight: float = 0.5,
        jaccard_weight: float = 0.5,
        log_loss: bool = False,
        log_sigmoid: bool = True,
        smooth: float = 0.0,
        reduction: str = "mean",
    ):

        # super(BCEJaccardLoss, self).__init__()
        self.bce_weight = bce_weight
        self.jaccard_weight = jaccard_weight
        self.bce_loss = BCEWithLogitsLoss(reduction=reduction)  # from logits
        self.log_sigmoid = log_sigmoid
        self.log_loss = log_loss
        self.smooth = smooth
        self.reduction = reduction

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        """
        Returns forward propagation of the loss

        Args:
            y_pred: binary masks of shape Nx1xHxW
            y_true: binary masks of shape Nx1xHxW
        """
        if y_true.size() != y_pred.size():
            raise ValueError(
                "Model outputs and targetd mush have the same size: Nx1xHxW"
            )

        loss: Tensor = self.bce_weight * self.bce_loss(y_pred, y_true)

        if self.jaccard_weight != 0:
            eps = 1e-7
            if self.log_sigmoid:
                # Log-Exp gives more numerically stable result and does not cause vanishing gradient on
                # extreme values 0 and 1
                outputs = F.logsigmoid(y_pred).exp()
            else:
                outputs = torch.sigmoid(y_pred)
            targets = (y_true == 1).type_as(outputs)

            intersection = torch.sum(outputs * targets)
            union = torch.sum(outputs + targets) - intersection
            jaccard_loss: Tensor = (intersection + self.smooth) / (
                union.clamp_min(eps) + self.smooth
            )

            if self.log_loss:
                jaccard_loss = -torch.log(jaccard_loss.clamp_min(eps))
            else:
                jaccard_loss = 1 - jaccard_loss

            if self.reduction == "mean":
                jaccard_loss = jaccard_loss.mean()
            elif self.reduction == "sum":
                jaccard_loss = jaccard_loss.sum()

            loss += self.jaccard_weight * jaccard_loss

        return loss


def test_loss() -> None:
    size = (2, 1, 4, 4)
    # uniform distrubution of number in -2 .. +2 range
    y_pred = (
        4 * torch.rand(size) - 2
    )  # uniform distrubution of number in -2 .. +2 range
    y_true = torch.randint(low=0, high=2, size=size).type_as(y_pred)
    print(f"y_pred {y_pred}")
    print(f"torch.sigmoid(y_pred) {torch.sigmoid(y_pred)}")
    print(f"y_true {y_true}")

    criterion = BCEWithLogitsLoss(reduction="mean")
    loss_bce = criterion(y_pred, y_true)
    print(f"BCEWithLogitsLoss {loss_bce}")

    criterion = BCEJaccardLoss(bce_weight=1, jaccard_weight=0, log_sigmoid=True)
    loss = criterion(y_pred, y_true)
    print(f"BCEJaccardLoss(bce_weight=1, jaccard_weight=0, log_sigmoid=True) {loss}")

    criterion = BCEJaccardLoss(bce_weight=0, jaccard_weight=1)
    loss = criterion(y_pred, y_true)
    print(f"BCEJaccardLoss(bce_weight=0, jaccard_weight=1, log_sigmoid=True) {loss}")

    criterion = BCEJaccardLoss(bce_weight=0.5, jaccard_weight=0.5, log_loss=False)
    loss = criterion(y_pred, y_true)
    print(
        f"BCEJaccardLoss(bce_weight=0.5, jaccard_weight=0.5, log_sigmoid=True) {loss}"
    )

    criterion = BCEJaccardLoss(bce_weight=0.5, jaccard_weight=0.5, log_loss=True)
    loss = criterion(y_pred, y_true)
    print(f"BCEJaccardLoss(bce_weight=0.5, jaccard_weight=0.5, log_loss=True) {loss}")


if __name__ == "__main__":
    test_loss()



"""
Soft Jaccard Loss

Adapted from:
    Source: https://github.com/BloodAxe/pytorch-toolbelt/blob/develop/pytorch_toolbelt/losses/
    Author: E. Khvedchenya (BloodAxe)
"""
import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn.modules.loss import _Loss

__all__ = ["JaccardLoss"]


class JaccardLoss(_Loss):
    """
    Implementation of soft Jaccard loss for binary image segmentation task
    calculated from logits
        Args:
            log_loss: if True, loss computed as `-log(jaccard)`; otherwise `1 - jaccard`
            log_sigmoid: if True, uses F.logsigmoid(y_pred).exp(), otherwise uses torch.sigmoid(y_pred)
            smooth: smooth the loss values, should be float between 0 and 1
            reduction: str = 'mean', 'sum' or 'none'. Default = 'mean'
        Output: loss as a float or Tensor (for reduction 'none')

    """

    def __init__(
        self,
        log_sigmoid: bool = True,
        log_loss: bool = False,
        smooth: float = 0.0,
        reduction: str = "mean",
    ):

        super(JaccardLoss, self).__init__()

        self.log_sigmoid = log_sigmoid
        self.log_loss = log_loss
        self.smooth = smooth
        self.reduction = reduction

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        """
        Args:
            y_pred: binary masks of shape Nx1xHxW
            y_true: binary masks of shape Nx1xHxW
        """
        if y_true.size() != y_pred.size():
            raise ValueError(
                "Model outputs and targetd mush have the same size: Nx1xHxW"
            )
        eps = 1e-7

        if self.log_sigmoid:
            # Log-Exp more numerically stable and does not cause vanishing gradient on
            # extreme values of 0 and 1
            outputs = F.logsigmoid(y_pred).exp()
        else:
            # outputs = y_pred
            outputs = torch.sigmoid(y_pred)
        targets = (y_true == 1).type_as(outputs)  # ensure the same type

        intersection = torch.sum(outputs * targets)
        union = torch.sum(outputs + targets) - intersection
        jaccard_loss = (intersection + self.smooth) / (
            union.clamp_min(eps) + self.smooth
        )

        if self.log_loss:
            jaccard_loss = -torch.log(jaccard_loss.clamp_min(eps))
        else:
            jaccard_loss = 1 - jaccard_loss

        if self.reduction == "mean":
            jaccard_loss = jaccard_loss.mean()
        elif self.reduction == "sum":
            jaccard_loss = jaccard_loss.sum()

        return jaccard_loss


def test_loss():
    size = (2, 1, 2, 2)
    # uniform distrubution of number in -2 .. +2 range
    y_pred = (
        4 * torch.rand(size) - 2
    )  # uniform distrubution of number in -2 .. +2 range
    y_true = torch.randint(low=0, high=2, size=size).type_as(y_pred)
    print(f"y_pred \n {y_pred}")
    print(f"torch.sigmoid(y_pred) \n {torch.sigmoid(y_pred)}")
    print(f"y_true \n {y_true}")

    criterion = JaccardLoss(log_sigmoid=True)
    loss = criterion(y_pred, y_true)
    print(f"JaccardLoss() {loss}")

    criterion = JaccardLoss(log_sigmoid=False)
    loss = criterion(y_true, y_true)
    print(f"(y_true, y_true) JaccardLoss(log_sigmoid=False) {loss}")


if __name__ == "__main__":
    test_loss()



import os
import random
from typing import Any, Dict, Tuple

import numpy as np
import torch
from torch.nn import Module
from torch.optim import Optimizer


def fix_seed(seed: int = 1234) -> None:
    """
    Fix all random seeds for reproducibility
    PyTorch
    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = True


def fix_seeds_tf(seed: int = 1234) -> None:
    """
    Fix all random seeds for reproducibility
    for Tensorflow
    """
    import tensorflow as tf

    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    os.environ["TF_DETERMINISTIC_OPS"] = str(seed)
    np.random.seed(seed)
    tf.random.set_seed(seed)


def load_optim(
    optimizer: Optimizer, checkpoint_path: str, device: torch.device
) -> Optimizer:
    """
    Load optimizer to continuer training
        Args:
            optimizer      : initialized optimizer
            checkpoint_path: path to the checkpoint
            device         : device to send optimizer to (must be the same as in the model)

        Note: must be called after initializing the model

        Output: optimizer with the loaded state
    """
    checkpoint = torch.load(checkpoint_path)
    optimizer.load_state_dict(checkpoint["optimizer"])
    for state in optimizer.state.values():
        for k, v in state.items():
            if torch.is_tensor(v):
                state[k] = v.to(device)

    for param_group in optimizer.param_groups:
        print("learning_rate: {}".format(param_group["lr"]))
    print("Loaded optimizer {} state from {}".format(optimizer, checkpoint_path))
    return optimizer


def save_ckpt(model: Module, optimizer: Optimizer, checkpoint_path: str) -> None:
    """
    Save model and optimizer checkpoint to continuer training
    """
    torch.save(
        {
            "model": model.state_dict(),
            "optimizer": optimizer.state_dict(),
        },
        checkpoint_path,
    )
    print("Saved model and optimizer state to {}".format(checkpoint_path))


def load_ckpt(checkpoint_path: str) -> Any:
    """
    Load checkpoint to continuer training
        Args:
            checkpoint_path: path to the checkpoint

        Output: (dict) 0f the checkpoint state

    """
    checkpoint = torch.load(checkpoint_path)
    return checkpoint


def load_model(model: Module, checkpoint_path: str) -> Tuple[Any, Dict[str, Any]]:
    """
    Load model weigths to continuer training
        Args:
            model          : nn model
            checkpoint_path: path to the checkpoint

        Output:
            (nn.Module) nn model with weights
            (dict) 0f the checkpoint state
    """
    checkpoint = torch.load(checkpoint_path)
    model.load_state_dict(checkpoint["model"])

    return model, checkpoint


def collate_fn(batch):
    return tuple(zip(*batch))


def load_weights(model: Module, weights_file: str):
    """Load weight to a torch model from file"""
    model.load_state_dict(torch.load(weights_file))
    return model



from typing import Any, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor

SMOOTH = 1e-7
NUM_CLASSES = 2


def binary_iou_numpy(output: Tensor, target: Tensor, from_logits: bool = True) -> float:
    """
    Batch images IoUs for a  binary segmentation
    Args:
        output: model output as a torch.Tensor
        target: true masks as a torch.Tensor
    """
    if from_logits:
        # Binary output
        output = torch.sigmoid(output)
        output.round()  # output.cuda().round()
    output = np.asarray(output.detach().cpu().numpy())
    target = np.asarray(target.detach().cpu().numpy())
    if output.ndim == 4:
        output = np.squeeze(output, axis=1)  # BATCH x 1 x H x W => BATCH x H x W
    if target.ndim == 4:
        target = np.squeeze(target, axis=1)
    # Remove background
    target = (target == 1)
    # Compute area intersection:
    intersection = (output * target).sum()
    # Union, sums ones excluding batch dimension
    union = output.sum() + target.sum() - intersection
    iou = intersection / (union + SMOOTH)
    iou_mean: float = iou.mean()

    return iou_mean


def binary_iou_pytorch(output: Tensor, target: Tensor, from_logits: bool = True) -> float:
    #  BATCH x 1 x H x W shape
    if output.ndim == 4:
        output = output.squeeze(1)  # BATCH x 1 x H x W => BATCH x H x W
    if target.ndim == 4:
        target = target.squeeze(1)
    # Remove background
    target = (target == 1).type_as(output)
    if from_logits:
        # Binary output
        output = torch.sigmoid(output)
        output.round()  # output.cuda().round()
    # Compute area intersection:
    intersection = torch.sum(output * target)
    # Union, sums ones excluding batch dimension
    union = torch.sum(output + target) - intersection
    iou = intersection / (union + SMOOTH)
    iou_mean: float = iou.mean()

    return iou_mean


def test_iou():
    output = torch.tensor([[0, 0, 0, 1], [0, 0, 0, 1], [1, 0, 0, 0], [0, 1, 0, 0]]).unsqueeze(0)
    target = torch.tensor([[0, 0, 1, 1], [0, 0, 1, 1], [1, 1, 0, 0], [1, 1, 0, 0]]).unsqueeze(0)
    print(output, target)
    iou: float = binary_iou_numpy(output, target)
    print(f'iou numpy: {iou}')
    assert abs(iou - 0.5) < 0.01
    iou_torch: float = binary_iou_pytorch(output, target)
    print(f'iou pytorch: {iou_torch}')
    assert abs(iou_torch - 0.5) < 0.01


def official_pixel_acc(output: Tensor, target: Tensor) -> Tuple[float, int, int]:
    """
    Official pixel accuracy
    as from
    https://github.com/CSAILVision/sceneparsing/blob/master/evaluationCode/utils_eval.py
    # This function takes the prediction and label of a single image, returns pixel-wise accuracy
    # To compute over many images do:
    # for i = range(Nimages):
    #	(pixel_accuracy[i], pixel_correct[i], pixel_labeled[i]) = pixelAccuracy(imPred[i], imLab[i])
    # mean_pixel_accuracy = 1.0 * np.sum(pixel_correct) / (np.spacing(1) + np.sum(pixel_labeled))
    """
    output = np.asarray(output)
    target = np.asarray(target)
    pixel_labeled: int = np.sum(target > 0)
    pixel_correct: int = np.sum((output == target) * (target > 0))
    pixel_accuracy: float = 1.0 * pixel_correct / pixel_labeled

    return (pixel_accuracy, pixel_correct, pixel_labeled)


def pixel_acc(output: Tensor, target: Tensor, ignore_index: int = 0) -> float:
    """Calculate pixel accuracy between model prediction and target
       Args:
           output (torch.tensor): model output of shape (B,C,H,W);
                each class in one channel (C == n_classes)
           target (torch.tensor): target tensor of shape (B,H,W);
                class objects encoded by unique values
    """
    mask = (target != ignore_index)
    # output = torch.argmax(output, dim=1)
    correct = (target == output)
    accuracy: float = (correct * mask).sum().float() / mask.sum()
    
    return accuracy


def pytorch_iou(output: torch.tensor, target: torch.tensor, num_class: int = NUM_CLASSES) -> Tuple[Any, Any]:
    """
    PyTorch metric from official metric as in
    https://github.com/CSAILVision/sceneparsing/blob/master/evaluationCode/utils_eval.py
    """
    # Remove classes from unlabeled pixels in gt image
    output = output * (target > 0)
    # Compute area intersection:
    intersection = output * (output == target)
    area_intersection = torch.histc(intersection, bins=num_class, min=1, max=num_class)
    # Compute area union
    area_pred = torch.histc(output, bins=num_class, min=1, max=num_class)
    area_lab = torch.histc(target, bins=num_class, min=1, max=num_class)
    area_union = area_pred + area_lab - area_intersection

    return (area_intersection, area_union)
    

def official_iou(output: torch.tensor, target: torch.tensor, numClass: int = NUM_CLASSES) -> Tuple[Any, Any]:
    """
    Official metric from 
    https://github.com/CSAILVision/sceneparsing/blob/master/evaluationCode/utils_eval.py
    
    This function takes the prediction and label of a single image, returns intersection and union areas for each class
    To compute over many images do:
    for i in range(Nimages): 
        (area_intersection[:,i], area_union[:,i]) = intersectionAndUnion(imPred[i], imLab[i])
    IoU = 1.0 * np.sum(area_intersection, axis=1) / np.sum(np.spacing(1)+area_union, axis=1)
    """
    # output = torch.argmax(output, dim=1)
    output = np.asarray(output.detach().cpu().numpy())
    target = np.asarray(target.detach().cpu().numpy())
    # Remove classes from unlabeled pixels in gt image.We should not penalize detections in unlabeled portions 
    output = output * (target > 0)
    # Compute area intersection
    intersection = output * (output == target)
    (area_intersection, _) = np.histogram(intersection, bins=numClass, range=(1, numClass))
    # Compute area union
    (area_pred, _) = np.histogram(output, bins=numClass, range=(1, numClass))
    (area_lab, _) = np.histogram(target, bins=numClass, range=(1, numClass))
    area_union = area_pred + area_lab - area_intersection
    
    return (area_intersection, area_union)    


def mean_iou(output: np.array, target: np.array) -> float:
    """Calculate mean IoU score averaged for a batch
       Args:
           output (torch.tensor): model output of shape (B,C,H,W);
                each class in one channel (C == n_classes)
           target (torch.tensor): target tensor of shape (B,H,W);
                class objects encoded by unique values 
           ignore_empty (bool): if True, classes which are not presented on target mask,
                and not predicted by `net` are not included to metric calculation        
    """
    n_classes = output.shape[1]
    # Prepare output
    # output = torch.argmax(output, dim=1)
    # Remove classes from unlabeled pixels in gt image
    output = output * (target > 0)
    # convert target to onehot BHWC
    target = F.one_hot(target, n_classes)
    output = F.one_hot(output, n_classes)    
    # intersection-over-union
    intersection = (output * target).sum(dim=(0, 1, 2)).float()
    union = (output + target).sum(dim=(0, 1, 2)).float() - intersection
    iou = (intersection + SMOOTH) / (union + SMOOTH)
    iou_mean: float = iou.mean()

    return iou_mean


def cm(pred: torch.tensor, true: torch.tensor) -> Any:
    """
    Confusion matrix
    """
    c = pred.shape[1]
    pred = pred.argmax(dim=1).view(-1)
    true = true.view(-1)
    mat = torch.zeros(c, c, dtype=torch.long)
    mat = mat.index_put_((true, pred), torch.tensor(1), accumulate=True)

    return mat.double() / mat.sum()


def cm_grad(pred: Tensor, true: Tensor) -> Any:
    """
    Gradient of confusion matrix
    """
    c = pred.shape[1]
    pred = pred.softmax(dim=1).view(-1, c)
    true = true.view(-1)
    mat = torch.zeros(c, c, dtype=pred.dtype).index_add(0, true, pred) 

    return mat.double() / mat.sum()


def iou_torch(pred: Tensor, true: Tensor) -> Any:
    """
    Intersection over Union using Confusion matrix
    PyTorch version
    does not work with the batch, single images only
    """
    c = pred.shape[1]
    # Remove classes from unlabeled pixels in gt image 
    pred = pred * (true > 0)
    pred = pred.view(-1)  # flatten predictions
    true = true.view(-1)
    mat = torch.zeros(c, c, dtype=torch.long)  # confusion matrix shape
    mat = mat.index_put_((true, pred), torch.tensor(1), accumulate=True)

    return mat.diag() / (mat.sum(0) + mat.sum(1) - mat.diag()).clamp(1e-8)


def precision_at(iou: np.array, threshold: float = 0.5) -> Tuple[int, int, int]:
    """ 
    Get true positive, false positive, false negative at iou threshold
    """
    matches = iou >= threshold
    true_positives = np.sum(matches, axis=1) == 1  # Correct objects
    false_positives = np.sum(matches, axis=0) == 0  # Missed objects
    false_negatives = np.sum(matches, axis=1) == 0  # Extra objects
    tp, fp, fn = np.sum(true_positives), np.sum(false_positives), np.sum(false_negatives)

    return tp, fp, fn
    

if __name__ == "__main__":
    test_iou()



import cv2
import numpy as np


def pad_x32(image: np.array) -> np.array:

    h, w = image.shape[:2]

    pad_h = np.ceil(h / 32) * 32 - h
    pad_w = np.ceil(w / 32) * 32 - w

    pad_h_top = int(np.floor(pad_h / 2))
    pad_h_bot = int(np.ceil(pad_h / 2))
    pad_w_top = int(np.floor(pad_w / 2))
    pad_w_bot = int(np.ceil(pad_w / 2))

    padding = ((pad_h_top, pad_h_bot), (pad_w_top, pad_w_bot), (0, 0))
    pad = padding[:2] if image.ndim == 2 else padding
    image = np.pad(image, pad, mode="constant", constant_values=0)

    return image


def load_image(file_name: str) -> np.array:
    """
    Helper loads image from file
        Args:
            file_name: (str) name of the image file

        Output: image as a numpy array
    """
    image = cv2.imread(file_name, cv2.IMREAD_COLOR)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    return image


def normalize_sar(
    img: np.array,
    mean: np.array = [0.485, 0.456, 0.406, 0.406],
    std: np.array = [0.229, 0.224, 0.225, 0.225],
    max_value: float = 92.88,
) -> np.array:
    """
    Normalize image data in 4 channels to 0-1 range,
    then applymenaand std as in ImageNet pretrain, or any other
    """
    mean = np.array(mean, dtype=np.float32)
    mean *= max_value
    std = np.array(std, dtype=np.float32)
    std *= max_value

    img = img.astype(np.float32)
    img = img - mean
    img = img / std

    return img


def preprocess_minmax(img: np.array) -> np.array:
    """
    Normalize image data to 98-2 percentiles

    """
    im_min = np.percentile(img, 2)
    im_max = np.percentile(img, 98)
    im_range = im_max - im_min
    # print(f'percentile 2 {im_min}, percentile 98 {im_max}, im_range {im_range}')

    # Normalise to the percentile
    img = img.astype(np.float32)
    img = (img - im_min) / im_range
    img = img.clip(0, 1)

    return img
